/*
    Copyright  1995-2001, The AROS Development Team. All rights reserved.
    $Id: newfontcontents.c,v 1.4 2007/04/26 17:52:41 piru Exp $
 
    Desc:
    Lang: English
*/

/****************************************************************************************/

#include <diskfont/diskfont.h>
#include <utility/tagitem.h>
#include <dos/dos.h>
#include <proto/dos.h>
#include <proto/exec.h>
#include <string.h>
#include <aros/debug.h>
#include <aros/macros.h>
#include "diskfont_intern.h"

/****************************************************************************************/

#define USE_EXALL 1

/* Test */
#define  GetByte(x)  ((x)[0])
#define  GetWord(x)  ((x)[1] | ((x)[0] << 8))
#define  GetPtr(x)   (((x)[0] << 24) | ((x)[1] << 16) | ((x)[2] << 8) | ((x)[3]))

struct AmiTagItem
{
	ULONG ti_Tag;
	ULONG ti_Data;
};

/****************************************************************************************/

static LONG HandleEntry(CONST_STRPTR name, CONST_STRPTR filename,
                        struct MinList *list, struct FontContentsHeader *fch);
static VOID CopyContents(struct MinList *list, APTR mem);
VOID FreeBuffers(struct MinList *list);
struct AmiTagItem *AmiNextTagItem(struct AmiTagItem **ti);

/****************************************************************************************/

struct contentsBuffer
{
	struct MinNode      node;
	struct FontContents fc;
};

/****************************************************************************************/

#define TFC(node) ((struct TFontContents *)(&node->fc))

/*****************************************************************************
 
    NAME */
#include <proto/diskfont.h>

AROS_LH2(struct FontContentsHeader *, NewFontContents,

         /*  SYNOPSIS */
         AROS_LHA(BPTR  , fontsLock, A0),
         AROS_LHA(CONST_STRPTR, fontName, A1),

         /*  LOCATION */
         struct Library *, DiskfontBase, 7, Diskfont)

/*  FUNCTION
 
    Create an array of FontContents entries describing the fonts related
    with 'fontName' -- this is those in the directory with the same name
    as 'fontName' without the ".font" suffix.
 
    INPUTS
 
    fontsLock  --  A lock on the FONTS: directory or another directory
                   containing the font file and associated directory
		   exists.
    fontName   --  The font name (with the ".font" suffix).
 
    RESULT
 
    Pointer to a struct FontContentsHeader describing the font or NULL
    if something went wrong.
 
    NOTES
 
    EXAMPLE
 
    BUGS
 
    SEE ALSO
 
    DisposeFontContents()
 
    INTERNALS
 
    The whole issue of fonts being Amiga executable files is a big mess.
    We'd better do something about it (define a new format?).
        Some code here should use a function similar to the one in
    ReadDiskFont() -- however, we do not want a struct TextFont *.
 
    HISTORY
 
    5.8.1999  SDuvan  partial implementation
 
*****************************************************************************/
{
	AROS_LIBFUNC_INIT

	BPTR   oldDir;
	BPTR   lock, otLock;
	STRPTR suffix;
	UBYTE  name[MAXFONTNAME];
	struct MinList contentsList;

	struct FontContentsHeader *ret = NULL;
	struct FontContentsHeader fch =
	    {
		    FCH_ID , 0
	    };
	LONG err = TRUE;

	(void) DiskfontBase;

	strcpy(name, fontName);
	suffix = strrchr(name, '.');

	if (suffix == NULL || stricmp(suffix, ".font") != 0)
		return NULL;

	strcpy(suffix, ".otag");

	NEWLIST((struct List *) &contentsList);
	oldDir = CurrentDir(fontsLock);

	/* otLock will be an indicator of whether there exists a .otag file */
	otLock = Lock(name, SHARED_LOCK);
	UnLock(otLock);

	/* Get the correct directory name */
	name[strlen(name) - sizeof(".otag") + 1] = '\0';

	lock = Lock(name, SHARED_LOCK); /* Lock font directory */
	if (lock)
	{

#if USE_EXALL

#define EAC_BUFFERSIZE 512
		struct ExAllControl *eac;
		struct ExAllData *ead;
		UBYTE ex_buffer[EAC_BUFFERSIZE];

		if ((eac = AllocDosObject(DOS_EXALLCONTROL, NULL)))
		{
			LONG more;

			(void) CurrentDir(lock);

			/*eac->eac_LastKey = 0;*/
			err = FALSE;

			do
			{
				more = ExAll(lock, (APTR) ex_buffer, EAC_BUFFERSIZE, ED_TYPE, eac);

				if ((!more) && (IoErr() != ERROR_NO_MORE_ENTRIES))
				{
					err = TRUE;
					break;
				}

				if (eac->eac_Entries == 0)
				{
					continue;
				}

				ead = (struct ExAllData *)ex_buffer;
				do
				{
					LONG rc;

					/* Skip directories */
					if (ead->ed_Type >= 0) continue;

					rc = HandleEntry(name, ead->ed_Name, &contentsList, &fch);
					if (rc < 0)
					{
						ExAllEnd(lock, (APTR) ex_buffer, EAC_BUFFERSIZE, ED_TYPE, eac);
						err = TRUE;
						more = 0;
						break;
					}

				} while ((ead = ead->ed_Next));

			} while (more);

			FreeDosObject(DOS_EXALLCONTROL, eac);

			CurrentDir(oldDir); /* Must currentdir to old dir before unlocking! */
		}
#else

#if #cpu(powerpc)
		struct FileInfoBlock    _fib, *fib = &_fib;
#else
		UBYTE _fib[sizeof(struct FileInfoBlock) + 3];
		struct FileInfoBlock    *fib = (APTR) ((((IPTR) _fib) + 3) & ~3);
#endif

		if (Examine(lock, fib) && fib->fib_DirEntryType >= 0) /* must be a directory */
		{
			CurrentDir(lock);

			err = FALSE;

			/* Loop through the files in this font's directory */
			while (ExNext(lock, fib))
			{
				LONG rc;

				/* Skip directories */
				if(fib->fib_DirEntryType >= 0)
					continue;

				rc = HandleEntry(name, fib->fib_FileName, &contentsList, &fch);
				if (rc < 0)
				{
					err = TRUE;
					break;
				}

			} /* while (there are files left in the directory) */

			if ((IoErr() != ERROR_NO_MORE_ENTRIES) && (IoErr() != 0))
			{
				err = TRUE;
			}

			CurrentDir(oldDir); /* Must currentdir to old dir before unlocking! */

		} /* if (we could examine the font's directory) */

#endif /* USE_EXALL */

		UnLock(lock);

	} /* if (lock) */

	if (!err)
	{
		/* Warning: FixFonts and similar depend on the buffer being allocated
		 * with AllocVec. They peek the data size from -4 and subtract 4.
		 * - Piru
		 */
		ret = (struct FontContentsHeader *)AllocVec(sizeof(struct FontContentsHeader) + fch.fch_NumEntries*sizeof(struct TFontContents), MEMF_PUBLIC | MEMF_CLEAR);
		if (ret)
		{
			ret->fch_NumEntries = fch.fch_NumEntries;
			ret->fch_FileID = otLock == 0 ? fch.fch_FileID : OFCH_ID;

			CopyContents(&contentsList,
			             ret + 1);
		}
	}

	FreeBuffers(&contentsList);

	CurrentDir(oldDir);

	return ret;

	AROS_LIBFUNC_EXIT

} /* NewFontContents */

/****************************************************************************************/

static LONG HandleEntry(CONST_STRPTR name, CONST_STRPTR filename,
                        struct MinList *list, struct FontContentsHeader *fch)
{
	BPTR                   fontSeg;
	ULONG                  segsize;
	UBYTE                 *fileDfh;  /* struct DiskFontHeader */
	struct contentsBuffer *cNode;

	fontSeg = LoadSeg(filename);
	if (!fontSeg)
	{
		return 0;
	}

	segsize = ((ULONG *)BADDR(fontSeg))[-1];

	/* Skip NextSegment and ReturnCode */
	fileDfh = ((UBYTE *)BADDR(fontSeg)) + 8;

	if (segsize < 8 + 14+2+2+4+32+20+2+1+1 || GetWord(fileDfh + 14) != DFH_ID) /* dfh_FileID */
	{
		UnLoadSeg(fontSeg);
		return 0;
	}

	cNode = AllocMem(sizeof(struct contentsBuffer),
	                 MEMF_PUBLIC | MEMF_CLEAR);
	if (!cNode)
	{
		UnLoadSeg(fontSeg);
		return -1;
	}

	strcpy(cNode->fc.fc_FileName, name);
	strcat(cNode->fc.fc_FileName, "/");
	strcat(cNode->fc.fc_FileName, filename);

	/* TODO: Factor out the code in diskfont_io.c that extracts
	         tags and make it a function... */

	/* Embedded tags? */
	if(GetByte(fileDfh+14+2+2+4+32+20+2) & FSF_TAGGED) /* dfh_TF.tf_Style */
	{
		struct AmiTagItem *ti = (struct AmiTagItem *)(*(ULONG *)(fileDfh+14+2+2)); /* dfh_TagList */
		struct AmiTagItem *tPtr;
		struct AmiTagItem *item;
		LONG   nTags = 1; /* Include TAG_DONE */
		LONG   i;	        /* Loop variable */

		/* We cannot use the standard function here as the
		   TagItems saved on disk may not correspond very
		   well to those on this particular architecture. */
		while(AmiNextTagItem(&ti) != NULL)
			nTags++;

		TFC(cNode)->tfc_TagCount = nTags;
		fch->fch_FileID = TFCH_ID;

		tPtr = (struct AmiTagItem *)((IPTR)&(TFC(cNode)->tfc_TagCount) + 2 - nTags*sizeof(struct AmiTagItem));

		ti = (struct AmiTagItem *)(*(ULONG *)(fileDfh+14+2+2)); /* dfh_TagList */

		i = 0;
		while((item = AmiNextTagItem(&ti)) != NULL)
		{
			tPtr[i].ti_Tag  = AROS_LONG2BE(item->ti_Tag);
			tPtr[i].ti_Data = AROS_LONG2BE(item->ti_Data);
			i++;
		}
		/* Add TAG_DONE tag, but no data (to avoid writing over the
		   TagCount) */
		tPtr[i].ti_Tag = TAG_DONE;

	} /* if(this was a tagged font) */

	cNode->fc.fc_YSize = GetWord(fileDfh+14+2+2+4+32+20);  /* dfh_TF.tf_YSize */
	cNode->fc.fc_Style = GetByte(fileDfh+14+2+2+4+32+20+2); /* dfh_TF.tf_Style */
	cNode->fc.fc_Flags = GetByte(fileDfh+14+2+2+4+32+20+2+1); /* dfh_TF.tf_Flags */

	cNode->fc.fc_Flags = (cNode->fc.fc_Flags & ~(FPF_REMOVED | FPF_ROMFONT)) | FPF_DISKFONT;

	ADDTAIL((struct List *)list, (struct Node *)cNode);

	fch->fch_NumEntries++;

	UnLoadSeg(fontSeg);

	return 0;
}

/****************************************************************************************/

VOID FreeBuffers(struct MinList *list)
{
	struct contentsBuffer *node, *temp;

	ForeachNodeSafe(list, node, temp)
	{
		FreeMem(node, sizeof(*node));
	}
	NEWLIST(list);
}

/****************************************************************************************/

static VOID CopyContents(struct MinList *list, APTR mem)
{
	struct contentsBuffer *temp;

	ForeachNode(list, (struct Node *)temp)
	{
		CopyMem(&temp->fc, mem, sizeof(struct FontContents));
		mem += sizeof(struct FontContents);
	}
}

/****************************************************************************************/

struct AmiTagItem *AmiNextTagItem(struct AmiTagItem **ti)
{
	if((*ti)->ti_Tag == TAG_DONE)
		return NULL;

	return (*ti)++;
}

/****************************************************************************************/
